@@ -71,7 +71,7 @@ async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, s
             remote_text = f'{remote_ip}:{remote_port}'
         local_addr = None if server_ip in ('127.0.0.1', '::1', None) else (server_ip, 0)
         reader_cipher, _ = await prepare_ciphers(cipher, reader, writer, server_side=False)
-        lproto, user, host_name, port, lbuf, rbuf = await proto.accept(protos, reader=reader, writer=writer, authtable=AuthTable(remote_ip, authtime), reader_cipher=reader_cipher, sock=writer.get_extra_info('socket'), **kwargs)
+        lproto, user, host_name, port, client_connected = await proto.accept(protos, reader=reader, writer=writer, authtable=AuthTable(remote_ip, authtime), reader_cipher=reader_cipher, sock=writer.get_extra_info('socket'), **kwargs)
         if host_name == 'echo':
             asyncio.ensure_future(lproto.channel(reader, writer, DUMMY, DUMMY))
         elif host_name == 'empty':
@@ -87,13 +87,12 @@ async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, s
                 raise Exception(f'Connection timeout {roption.bind}')
             try:
                 reader_remote, writer_remote = await roption.prepare_connection(reader_remote, writer_remote, host_name, port)
-                writer.write(lbuf)
-                writer_remote.write(rbuf)
+                use_http = (await client_connected(writer_remote)) if client_connected else None
             except Exception:
                 writer_remote.close()
                 raise Exception('Unknown remote protocol')
             m = modstat(user, remote_ip, host_name)
-            lchannel = lproto.http_channel if rbuf else lproto.channel
+            lchannel = lproto.http_channel if use_http else lproto.channel
             asyncio.ensure_future(lproto.channel(reader_remote, writer, m(2+roption.direct), m(4+roption.direct)))
             asyncio.ensure_future(lchannel(reader, writer_remote, m(roption.direct), roption.connection_change))
     except Exception as ex:
@@ -304,6 +303,126 @@ class ProxySimple(ProxyDirect):
         else:
             return asyncio.start_server(handler, host=self.host_name, port=self.port, reuse_port=args.get('ruport'))
 
+class ProxyH2(ProxySimple):
+    def __init__(self, sslserver, sslclient, **kw):
+        super().__init__(sslserver=None, sslclient=None, **kw)
+        self.handshake = None
+        self.h2sslserver = sslserver
+        self.h2sslclient = sslclient
+    async def handler(self, reader, writer, client_side=True, stream_handler=None, **kw):
+        import h2.connection, h2.config, h2.events
+        reader, writer = proto.sslwrap(reader, writer, self.h2sslclient if client_side else self.h2sslserver, not client_side, None)
+        config = h2.config.H2Configuration(client_side=client_side)
+        conn = h2.connection.H2Connection(config=config)
+        streams = {}
+        conn.initiate_connection()
+        writer.write(conn.data_to_send())
+        while not reader.at_eof() and not writer.is_closing():
+            try:
+                data = await reader.read(65636)
+                if not data:
+                    break
+                events = conn.receive_data(data)
+            except Exception:
+                pass
+            writer.write(conn.data_to_send())
+            for event in events:
+                if isinstance(event, h2.events.RequestReceived) and not client_side:
+                    if event.stream_id not in streams:
+                        stream_reader, stream_writer = self.get_stream(conn, writer, event.stream_id)
+                        streams[event.stream_id] = (stream_reader, stream_writer)
+                        asyncio.ensure_future(stream_handler(stream_reader, stream_writer))
+                    else:
+                        stream_reader, stream_writer = streams[event.stream_id]
+                    stream_writer.headers.set_result(event.headers)
+                elif isinstance(event, h2.events.SettingsAcknowledged) and client_side:
+                    self.handshake.set_result((conn, streams, writer))
+                elif isinstance(event, h2.events.DataReceived):
+                    stream_reader, stream_writer = streams[event.stream_id]
+                    stream_reader.feed_data(event.data)
+                    conn.acknowledge_received_data(len(event.data), event.stream_id)
+                    writer.write(conn.data_to_send())
+                elif isinstance(event, h2.events.StreamEnded) or isinstance(event, h2.events.StreamReset):
+                    stream_reader, stream_writer = streams[event.stream_id]
+                    stream_reader.feed_eof()
+                    if not stream_writer.closed:
+                        stream_writer.close()
+                elif isinstance(event, h2.events.ConnectionTerminated):
+                    break
+                elif isinstance(event, h2.events.WindowUpdated):
+                    if event.stream_id in streams:
+                        stream_reader, stream_writer = streams[event.stream_id]
+                        stream_writer.window_update()
+        writer.write(conn.data_to_send())
+        writer.close()
+    def get_stream(self, conn, writer, stream_id):
+        reader = asyncio.StreamReader()
+        write_buffer = bytearray()
+        write_wait = asyncio.Event()
+        write_full = asyncio.Event()
+        class StreamWriter():
+            def __init__(self):
+                self.closed = False
+                self.headers = asyncio.get_event_loop().create_future()
+            def get_extra_info(self, key):
+                return writer.get_extra_info(key)
+            def write(self, data):
+                write_buffer.extend(data)
+                write_wait.set()
+            def drain(self):
+                writer.write(conn.data_to_send())
+                return writer.drain()
+            def is_closing(self):
+                return self.closed
+            def close(self):
+                self.closed = True
+                write_wait.set()
+            def window_update(self):
+                write_full.set()
+            def send_headers(self, headers):
+                conn.send_headers(stream_id, headers)
+                writer.write(conn.data_to_send())
+        stream_writer = StreamWriter()
+        async def write_job():
+            while not stream_writer.closed:
+                while len(write_buffer) > 0:
+                    while conn.local_flow_control_window(stream_id) <= 0:
+                        write_full.clear()
+                        await write_full.wait()
+                        if stream_writer.closed:
+                            break
+                    chunk_size = min(conn.local_flow_control_window(stream_id), len(write_buffer), conn.max_outbound_frame_size)
+                    conn.send_data(stream_id, write_buffer[:chunk_size])
+                    writer.write(conn.data_to_send())
+                    del write_buffer[:chunk_size]
+                if not stream_writer.closed:
+                    write_wait.clear()
+                    await write_wait.wait()
+            conn.send_data(stream_id, b'', end_stream=True)
+            writer.write(conn.data_to_send())
+        asyncio.ensure_future(write_job())
+        return reader, stream_writer
+    async def wait_h2_connection(self, local_addr, family):
+        if self.handshake is not None:
+            if not self.handshake.done():
+                await self.handshake
+        else:
+            self.handshake = asyncio.get_event_loop().create_future()
+            reader, writer = await super().wait_open_connection(None, None, local_addr, family)
+            asyncio.ensure_future(self.handler(reader, writer))
+            await self.handshake
+        return self.handshake.result()
+    async def wait_open_connection(self, host, port, local_addr, family):
+        conn, streams, writer = await self.wait_h2_connection(local_addr, family)
+        stream_id = conn.get_next_available_stream_id()
+        conn._begin_new_stream(stream_id, stream_id%2)
+        stream_reader, stream_writer = self.get_stream(conn, writer, stream_id)
+        streams[stream_id] = (stream_reader, stream_writer)
+        return stream_reader, stream_writer
+    def start_server(self, args, stream_handler=stream_handler):
+        handler = functools.partial(stream_handler, **vars(self), **args)
+        return super().start_server(args, functools.partial(self.handler, client_side=False, stream_handler=handler))
+
 class ProxyQUIC(ProxySimple):
     def __init__(self, quicserver, quicclient, **kw):
         super().__init__(**kw)
@@ -389,6 +508,93 @@ class ProxyQUIC(ProxySimple):
             asyncio.ensure_future(stream_handler(reader, writer, **vars(self), **args))
         return aioquic.asyncio.serve(self.host_name, self.port, configuration=self.quicserver, stream_handler=handler)
 
+class ProxyH3(ProxyQUIC):
+    def get_stream(self, conn, stream_id):
+        remote_addr = conn._quic._network_paths[0].addr
+        reader = asyncio.StreamReader()
+        class StreamWriter():
+            def __init__(self):
+                self.closed = False
+                self.headers = asyncio.get_event_loop().create_future()
+            def get_extra_info(self, key):
+                return dict(peername=remote_addr, sockname=remote_addr).get(key)
+            def write(self, data):
+                conn.http.send_data(stream_id, data, False)
+                conn.transmit()
+            async def drain(self):
+                conn.transmit()
+            def is_closing(self):
+                return self.closed
+            def close(self):
+                if not self.closed:
+                    conn.http.send_data(stream_id, b'', True)
+                    conn.transmit()
+                    conn.close_stream(stream_id)
+                self.closed = True
+            def send_headers(self, headers):
+                conn.http.send_headers(stream_id, [(i.encode(), j.encode()) for i, j in headers])
+                conn.transmit()
+        return reader, StreamWriter()
+    def get_protocol(self, server_side=False, handler=None):
+        import aioquic.asyncio, aioquic.quic.events, aioquic.h3.connection, aioquic.h3.events
+        class Protocol(aioquic.asyncio.QuicConnectionProtocol):
+            def __init__(s, *args, **kw):
+                super().__init__(*args, **kw)
+                s.http = aioquic.h3.connection.H3Connection(s._quic)
+                s.streams = {}
+            def quic_event_received(s, event):
+                if not server_side:
+                    if isinstance(event, aioquic.quic.events.HandshakeCompleted):
+                        self.handshake.set_result(s)
+                    elif isinstance(event, aioquic.quic.events.ConnectionTerminated):
+                        self.handshake = None
+                        self.quic_egress_acm = None
+                if s.http is not None:
+                    for http_event in s.http.handle_event(event):
+                        s.http_event_received(http_event)
+            def http_event_received(s, event):
+                if isinstance(event, aioquic.h3.events.HeadersReceived):
+                    if event.stream_id not in s.streams and server_side:
+                        reader, writer = s.create_stream(event.stream_id)
+                        writer.headers.set_result(event.headers)
+                        asyncio.ensure_future(handler(reader, writer))
+                elif isinstance(event, aioquic.h3.events.DataReceived) and event.stream_id in s.streams:
+                    reader, writer = s.streams[event.stream_id]
+                    if event.data:
+                        reader.feed_data(event.data)
+                    if event.stream_ended:
+                        reader.feed_eof()
+                    s.close_stream(event.stream_id)
+            def create_stream(s, stream_id=None):
+                if stream_id is None:
+                    stream_id = s._quic.get_next_available_stream_id(False)
+                    s._quic._get_or_create_stream_for_send(stream_id)
+                reader, writer = self.get_stream(s, stream_id)
+                s.streams[stream_id] = (reader, writer)
+                return reader, writer
+            def close_stream(s, stream_id):
+                if stream_id in s.streams:
+                    reader, writer = s.streams[stream_id]
+                    if reader.at_eof() and writer.is_closing():
+                        s.streams.pop(stream_id)
+        return Protocol
+    async def wait_h3_connection(self):
+        if self.handshake is not None:
+            if not self.handshake.done():
+                await self.handshake
+        else:
+            import aioquic.asyncio
+            self.handshake = asyncio.get_event_loop().create_future()
+            self.quic_egress_acm = aioquic.asyncio.connect(self.host_name, self.port, create_protocol=self.get_protocol(), configuration=self.quicclient)
+            conn = await self.quic_egress_acm.__aenter__()
+            await self.handshake
+    async def wait_open_connection(self, *args):
+        await self.wait_h3_connection()
+        return self.handshake.result().create_stream()
+    def start_server(self, args, stream_handler=stream_handler):
+        import aioquic.asyncio
+        return aioquic.asyncio.serve(self.host_name, self.port, configuration=self.quicserver, create_protocol=self.get_protocol(True, functools.partial(stream_handler, **vars(self), **args)))
+
 class ProxySSH(ProxySimple):
     def __init__(self, **kw):
         super().__init__(**kw)
@@ -544,11 +750,14 @@ def proxies_by_uri(uri_jumps):
         jump = proxy_by_uri(uri, jump)
     return jump
 
+sslcontexts = []
+
 def proxy_by_uri(uri, jump):
     scheme, _, uri = uri.partition('://')
     url = urllib.parse.urlparse('s://'+uri)
     rawprotos = [i.lower() for i in scheme.split('+')]
     err_str, protos = proto.get_protos(rawprotos)
+    protonames = [i.name for i in protos]
     if err_str:
         raise argparse.ArgumentTypeError(err_str)
     if 'ssl' in rawprotos or 'secure' in rawprotos:
@@ -558,18 +767,25 @@ def proxy_by_uri(uri, jump):
         if 'ssl' in rawprotos:
             sslclient.check_hostname = False
             sslclient.verify_mode = ssl.CERT_NONE
+        sslcontexts.append(sslserver)
+        sslcontexts.append(sslclient)
     else:
         sslserver = sslclient = None
-    if 'quic' in rawprotos:
+    if 'quic' in rawprotos or 'h3' in protonames:
         try:
             import ssl, aioquic.quic.configuration
         except Exception:
             raise Exception('Missing library: "pip3 install aioquic"')
-        import logging
-        quicserver = aioquic.quic.configuration.QuicConfiguration(is_client=False)
-        quicclient = aioquic.quic.configuration.QuicConfiguration()
+        quicserver = aioquic.quic.configuration.QuicConfiguration(is_client=False, max_stream_data=2**60, max_data=2**60, idle_timeout=SOCKET_TIMEOUT)
+        quicclient = aioquic.quic.configuration.QuicConfiguration(max_stream_data=2**60, max_data=2**60, idle_timeout=SOCKET_TIMEOUT*5)
         quicclient.verify_mode = ssl.CERT_NONE
-    protonames = [i.name for i in protos]
+        sslcontexts.append(quicserver)
+        sslcontexts.append(quicclient)
+    if 'h2' in rawprotos:
+        try:
+            import h2
+        except Exception:
+            raise Exception('Missing library: "pip3 install h2"')
     urlpath, _, plugins = url.path.partition(',')
     urlpath, _, lbind = urlpath.partition('@')
     plugins = plugins.split(',') if plugins else None
@@ -611,6 +827,10 @@ def proxy_by_uri(uri, jump):
                       host_name=host_name, port=port, unix=not loc, lbind=lbind, sslclient=sslclient, sslserver=sslserver)
         if 'quic' in rawprotos:
             proxy = ProxyQUIC(quicserver, quicclient, **params)
+        elif 'h3' in protonames:
+            proxy = ProxyH3(quicserver, quicclient, **params)
+        elif 'h2' in protonames:
+            proxy = ProxyH2(**params)
         elif 'ssh' in protonames:
             proxy = ProxySSH(**params)
         else:
@@ -646,7 +866,7 @@ async def test_url(url, rserver):
         print(headers.decode()[:-4])
         print(f'--------------------------------')
         body = bytearray()
-        while 1:
+        while not reader.at_eof():
             s = await reader.read(65536)
             if not s:
                 break
@@ -677,15 +897,8 @@ def main():
     args = parser.parse_args()
     if args.sslfile:
         sslfile = args.sslfile.split(',')
-        for option in args.listen:
-            if option.sslclient:
-                option.sslclient.load_cert_chain(*sslfile)
-                option.sslserver.load_cert_chain(*sslfile)
-        for option in args.listen+args.ulisten+args.rserver+args.urserver:
-            if isinstance(option, ProxyQUIC):
-                option.quicserver.load_cert_chain(*sslfile)
-            if isinstance(option, ProxyBackward) and isinstance(option.backward, ProxyQUIC):
-                option.backward.quicserver.load_cert_chain(*sslfile)
+        for context in sslcontexts:
+            context.load_cert_chain(*sslfile)
     elif any(map(lambda o: o.sslclient or isinstance(o, ProxyQUIC), args.listen+args.ulisten)):
         print('You must specify --ssl to listen in ssl mode')
         return
